from mmc.plugins.base import ldapUserGroupControl
from mmc.support.config import PluginConfig

import ldap.modlist

import os
import pwd
import grp

VERSION = "0.6.2"
APIVERSION = "0.0.2"
REVISION = ""

SIP_CONFIG_FILE = '/etc/asterisk/sip.conf'
EXTENSIONS_CONFIG_FILE = '/etc/asterisk/extensions.conf'
TRUNK_SIP_CONFIG_FILE = '/etc/asterisk/trunk_sip.conf'
DIALPLAN_CONFIG_FILE = '/etc/asterisk/dialplan.conf'


def getVersion():
    return VERSION


def getApiVersion():
    return APIVERSION


def getRevision():
    return REVISION


def activate():
    return True


def get_users(filter=""):
    if not filter:
        filter = "*"
    return UserAsterisk().get_users(filter=filter)


def has_object_class(uid):
    return UserAsterisk().has_object_class(uid)


def add_object_class(uid):
    UserAsterisk().add_object_class(uid)


def del_object_class(uid):
    UserAsterisk().del_object_class(uid)


def get_phone_number(uid):
    return UserAsterisk().get_phone_number(uid)


def update_phone_number(uid, phone_number):
    if UserAsterisk().phone_number_exists(phone_number):
        raise PhoneNumberAlreadyTaken()
    old_phone_number = UserAsterisk().get_phone_number(uid)

    # if the user had phone number and is different than the new phone number
    if old_phone_number and old_phone_number != phone_number:
        UserAsterisk().del_phone_extension(old_phone_number)

    UserAsterisk().update_phone_number(uid, phone_number)

    if phone_number:
        UserAsterisk().add_phone_extension(uid, phone_number)


def update_phone_password(uid, password):
    if password:
        temp_realm = os.popen('grep "^realm=" {}'.format(SIP_CONFIG_FILE))
        realm = temp_realm.read().split('realm=')[1].strip('\n')

        temp_password = os.popen('echo -n "' + uid + ':' + realm + ':' +
                                 password + '" | md5sum | cut -c1-32')

        password = "{md5}" + temp_password.read().strip('\n')

    UserAsterisk().update_phone_password(uid, password)


def has_voicemail(uid):
    return UserAsterisk().has_voicemail(uid)


def get_voicemail_attributes(uid):
    return UserAsterisk().get_voicemail_attributes(uid)


def update_voicemail(uid, phone_number, voicemail, password):
    if phone_number:
        if UserAsterisk().has_voicemail(uid):
            UserAsterisk().del_voicemail_extension(phone_number)

        UserAsterisk().update_voicemail(uid, phone_number)
        UserAsterisk().add_voicemail_extension(uid, phone_number,
                                               voicemail, password)

    else:
        UserAsterisk().del_voicemail_extension(
            UserAsterisk().get_phone_number(uid))

        UserAsterisk().update_voicemail(uid, "")


def get_voicemail_phone_number():
    return UserAsterisk().get_voicemail_phone_number()


def update_voicemail_phone_number(phone_number):
    if UserAsterisk().get_voicemail_phone_number():
        UserAsterisk().del_voicemail_phone_number(
            UserAsterisk().get_voicemail_phone_number())

    if phone_number:
        UserAsterisk().add_voicemail_phone_number(phone_number)


def get_groups(filter=""):
    if not filter:
        filter = "*"
    return UserAsterisk().get_groups(filter=filter)


def get_group(group):
    return UserAsterisk().get_group(group)


def update_group(name, old_phone_number, phone_number, list_users):
    if UserAsterisk().phone_number_exists(phone_number):
        raise PhoneNumberAlreadyTaken()
    if name:
        if UserAsterisk().group_exists(name) or old_phone_number:
            UserAsterisk().del_group_extension(old_phone_number)

        UserAsterisk().add_group_extension(name, phone_number, list_users)

    else:
        UserAsterisk().del_group_extension(old_phone_number)


def get_sip_config():
    return TrunkSIPConfig().get_sip_config()


def update_sip_config(sip_config):
    TrunkSIPConfig().update_sip_config(sip_config)


def get_dialplan_config():
    return TrunkSIPConfig().get_dialplan_config()


def update_dialplan_config(dialplan_config):
    TrunkSIPConfig().update_dialplan_config(dialplan_config)


class AsteriskException(Exception):

    def __init__(self, message):
        Exception.__init__(self)
        self.message = message

    def __str__(self):
        return self.message


class PhoneNumberAlreadyTaken(AsteriskException):

    def __init__(self):
        AsteriskException.__init__(self, "Phone number is already taken")


class UserAsterisk(ldapUserGroupControl):
    def __init__(self, conf_file_base=None):
        ldapUserGroupControl.__init__(self, conf_file_base)
        self.conf = PluginConfig("asterisk", None)
        self.extensionsDN = self.conf.get('main', 'extensionsOU')

    def get_users(self, filter="*"):
        """
        Return the list of users that have asterisk enabled

        @param filter: filter by uid or phone number
        @type filter: str
        """
        search_filter = '(&(objectClass=AsteriskSIPUser)(|(uid=%(filter)s)(AstAccountCallerID=%(filter)s)))' % {'filter': filter}
        return self.searchUserAdvance(search_filter)

    def has_object_class(self, uid):
        """
        Return True if the user has the asterisk object class

        @param uid: user uid
        @type uid: str
        """

        if uid == "":
            raise ValueError('The uid cannot be empty.')

        list_attributes = self.getDetailedUser(uid)["objectClass"]
        has_object_class = "AsteriskSIPUser" in list_attributes
        return has_object_class

    def add_object_class(self, uid):
        """
        Add the asterisk object class to the user
        and fill other static attributes

        @param uid: user uid
        @type uid: str
        """
        list_attributes = self.getDetailedUser(uid)["objectClass"]

        list_attributes.append("AsteriskSIPUser")
        self.changeUserAttributes(uid, "objectClass", list_attributes)
        self.changeUserAttributes(uid, "AstAccountAllowedCodec", "alaw")
        self.changeUserAttributes(uid, "AstAccountCanReinvite", "yes")
        self.changeUserAttributes(uid, "AstAccountContext", "internal")
        self.changeUserAttributes(uid, "AstAccountHost", "dynamic")
        self.changeUserAttributes(uid, "AstAccountNAT", "force_rport")
        self.changeUserAttributes(uid, "AstAccountQualify", "yes")
        self.changeUserAttributes(uid, "AstAccountType", "friend")
        self.changeUserAttributes(uid, "AstAccountDefaultUser", uid)

    def del_object_class(self, uid):
        """
        Remove the asterisk object class to the user
        and all asterisk attributes

        @param uid: user uid
        @type uid: str
        """
        self.changeUserAttributes(uid, "AstAccountAllowedCodec", "")
        self.changeUserAttributes(uid, "AstAccountCanReinvite", "")
        self.changeUserAttributes(uid, "AstAccountContext", "")
        self.changeUserAttributes(uid, "AstAccountHost", "")
        self.changeUserAttributes(uid, "AstAccountNAT", "")
        self.changeUserAttributes(uid, "AstAccountQualify", "")
        self.changeUserAttributes(uid, "AstAccountType", "")
        self.changeUserAttributes(uid, "AstAccountDefaultUser", "")

        if "AstAccountLastQualifyMilliseconds" in self.getDetailedUser(uid):
            self.changeUserAttributes(uid, "AstAccountLastQualifyMilliseconds", "")

        list_attributes = self.getDetailedUser(uid)["objectClass"]

        list_attributes.remove("AsteriskSIPUser")
        self.changeUserAttributes(uid, "objectClass", list_attributes)

    def get_phone_number(self, uid):
        """
        Return the phone number for a user

        @param uid: user uid
        @type uid: str
        """

        phone_number = ""
        if "AstAccountCallerID" in self.getDetailedUser(uid):
            phone_number = self.getDetailedUser(uid)["AstAccountCallerID"][0]

        return phone_number

    def update_phone_number(self, uid, phone_number):
        """
        Update the AstAccountCallerID attribute

        @param uid: user uid
        @type uid: str

        @param phone_number: user phone number
        @type phone_number: str
        """
        if self.phone_number_exists(phone_number):
            raise PhoneNumberAlreadyTaken()
        self.changeUserAttributes(uid, "AstAccountCallerID", phone_number)

    def update_phone_password(self, uid, password):
        """
        Set the AstAccountRealmedPassword attribute

        @param uid: user uid
        @type uid: str

        @param password: user password
        @type password: str
        """
        self.changeUserAttributes(uid, "AstAccountRealmedPassword", password)

    def has_voicemail(self, uid):
        """
        Return True if the user has voicemail configured7

        @param uid: user uid
        @type uid: str
        """

        has_voicemail = "AstAccountMailbox" in self.getDetailedUser(uid)

        return has_voicemail

    def get_voicemail_attributes(self, uid):
        """
        Return voicemail email and password for an asterisk user

        @param uid: user uid
        @type uid: str
        """

        attributes = ['', '']
        phone_number = self.getDetailedUser(uid)["AstAccountCallerID"][0]

        basedn = self.conf.get('main', 'voicemailOU')

        results = self.l.search_s(basedn, ldap.SCOPE_SUBTREE,
                                  '(sn=' + phone_number + '*)',
                                  ['AstVoicemailEmail', 'AstVoicemailPassword'])

        if len(results) == 1:
            attributes = [results[0][1].get('AstVoicemailEmail')[0],
                          results[0][1].get('AstVoicemailPassword')[0]]

        return attributes

    def update_voicemail(self, uid, phone_number):
        """
        Enable a voicemail for this user using the supplied password

        @param uid: user uid
        @type uid: str

        @param phone_number: user phone number
        @type phone_number: str
        """

        self.changeUserAttributes(uid, "AstAccountMailbox", phone_number)

    def get_voicemail_phone_number(self):
        """
        Return the voicemail phone number which called by asterisk users
        to listen their voicemails
        """

        phone_number = ""

        results = self.l.search_s(self.extensionsDN, ldap.SCOPE_SUBTREE,
                                  '(AstApplication=VoicemailMain)',
                                  ['AstExtension'])

        if len(results) > 0:
            phone_number = results[0][1].get('AstExtension')[0]

        return phone_number

    def add_voicemail_phone_number(self, phone_number):
        """
        Add the voicemail phone number in asterisk extensions

        @param phone_number: voicemail phone number
        @type phone_number: str
        """

        dn = "cn=" + phone_number + "-1, " + self.extensionsDN
        entry = {
            "cn": phone_number + "-1",
            "sn": phone_number + "-1",
            "AstApplication": "VoicemailMain",
            "AstContext": "internal",
            "AstPriority": "1",
            "AstApplicationData": phone_number + ",u",
            "AstExtension": phone_number,
            "objectClass": ("top", "inetOrgPerson", "AsteriskExtension")
        }
        modlist = ldap.modlist.addModlist(entry)
        self.l.add_s(dn, modlist)

    def del_voicemail_phone_number(self, phone_number):
        """
        Delete the voicemail phone number in asterisk extensions

        @param phone_number: voicemail phone number
        @type phone_number: str
        """

        dn = "cn=" + phone_number + "-1, " + self.extensionsDN
        self.delRecursiveEntry(dn)

    def add_phone_extension(self, uid, phone_number):
        """
        Add an asterisk extension entry in directory

        @param uid: username
        @type uid: str

        @param phone_number: user phone number
        @type phone_number: str
        """

        if self.phone_number_exists(phone_number):
            raise PhoneNumberAlreadyTaken

        dn = "cn=" + phone_number + "-1, " + self.extensionsDN
        entry = {
            "cn": phone_number + "-1",
            "sn": phone_number + "-1",
            "AstApplication": "Dial",
            "AstContext": "internal",
            "AstExtension": phone_number,
            "AstPriority": "1",
            "objectClass": ("top", "inetOrgPerson", "AsteriskExtension"),
            "AstApplicationData": "SIP/" + uid + ",20"
        }

        modlist = ldap.modlist.addModlist(entry)
        self.l.add_s(dn, modlist)

    def del_phone_extension(self, phone_number):
        """
        Delete an asterisk extension entry in directory

        @param phone_number: user phone number
        @type phone_number: str
        """

        dn = "cn=" + phone_number + "-1, " + self.extensionsDN
        self.delRecursiveEntry(dn)

    def add_voicemail_extension(self, uid, phone_number, voicemail, password):
        """
        Add an asterisk extension entry and
        a asterisk voicemail entry in directory

        @param uid: username
        @type uid: str

        @param phone_number: user phone number
        @type phone_number: str

        @param voicemail: user mail
        @type voicemail: str

        @param password: user mail password
        @type password: str
        """

        dn = "cn=" + phone_number + "-2, " + self.extensionsDN
        entry = {
            "cn": phone_number + "-2",
            "sn": phone_number + "-2",
            "AstApplication": "Voicemail",
            "AstContext": "internal",
            "AstPriority": "2",
            "AstApplicationData": phone_number + ",u",
            "AstExtension": phone_number,
            "objectClass": ("top", "inetOrgPerson", "AsteriskExtension")
        }
        modlist = ldap.modlist.addModlist(entry)
        self.l.add_s(dn, modlist)

        dn = "cn=" + phone_number + ", " + self.conf.get("main", "voicemailOU")
        entry = {
            "cn": phone_number,
            "sn": phone_number,
            "AstContext": "internal",
            "AstVoicemailMailbox": phone_number,
            "AstVoicemailPassword": password,
            "AstVoicemailEmail": voicemail,
            "AstVoicemailFullname": self.getDetailedUser(uid)["cn"][0],
            "objectClass": ("top", "inetOrgPerson", "AsteriskVoicemail")
        }
        modlist = ldap.modlist.addModlist(entry)
        self.l.add_s(dn, modlist)

    def del_voicemail_extension(self, phone_number):
        """
        Del an asterisk extension entry and
        a asterisk voicemail entry in directory

        @param phone_number: user phone number
        @type phone_number: str
        """

        dn = "cn=" + phone_number + "-2, " + self.extensionsDN
        self.delRecursiveEntry(dn)

        dn = "cn=" + phone_number + ", " + self.conf.get("main", "voicemailOU")
        self.delRecursiveEntry(dn)

    def group_exists(self, name):
        """
        Return True if the group exist, False if not

        @param name: group name
        @type name: str
        """
        results = self.l.search_s(self.extensionsDN, ldap.SCOPE_SUBTREE, '(sn=group_' + name + ')')
        if len(results) == 1:
            return True
        else:
            return False

    def phone_number_exists(self, number):
        """
        Return True if the number exists, False if not

        @param number: phone number
        @type number: str
        """
        if self.l.search_s(self.extensionsDN, ldap.SCOPE_SUBTREE, '(AstExtension=%s)' % number):
            return True
        else:
            return False

    def get_groups(self, filter='*'):
        """Return list of asterisk caller groups"""

        group_list = []

        results = self.l.search_s(self.extensionsDN, ldap.SCOPE_SUBTREE,
                                  '(|(sn=group_%(filter)s)(AstExtension=%(filter)s))' % {'filter': filter},
                                  ['AstExtension', 'AstApplicationData', 'sn'])

        mapping = [('SIP', ''), ('/', ''), (',', ''), ('20', '')]

        if len(results) > 0:
            for group in results:
                if not group[1]['sn'][0].startswith('group_'):
                    continue
                name = group[1].get('sn')[0].strip('group_')
                ast_application_data = group[1].get('AstApplicationData')[0]

                for k, v in mapping:
                    ast_application_data = ast_application_data.replace(k, v)

                user_list = ast_application_data.split('&')
                group_list.append([name, group[1].get('AstExtension')[0],
                                   user_list])

        return group_list

    def get_group(self, group):
        """ Return asterisk group infos
        """
        groups = self.get_groups(filter=group)
        if groups:
            return groups[0]
        return None

    def add_group_extension(self, name, phone_number, list_users):
        """
        Add an asterisk group extension entry

        @param name: group name
        @type name: str

        @param phone_number: group phone number
        @type phone_number: str

        @param list_users: list of users's uid
        @type list_users: list str []
        """

        if self.phone_number_exists(phone_number):
            raise PhoneNumberAlreadyTaken

        application_data = []

        for user_uid in list_users:
            application_data.append('SIP/' + user_uid)

        dn = "cn=" + phone_number + "-1, " + self.extensionsDN
        entry = {
            "cn": phone_number + "-1",
            "sn": "group_" + name,
            "AstApplication": "Dial",
            "AstContext": "internal",
            "AstExtension": phone_number,
            "AstPriority": "1",
            "objectClass": ("top", "inetOrgPerson", "AsteriskExtension"),
            "AstApplicationData": '&'.join(application_data) + ",20"
        }

        modlist = ldap.modlist.addModlist(entry)
        self.l.add_s(dn, modlist)

        dn = "cn=" + phone_number + "-2, " + self.extensionsDN
        entry = {
            "cn": phone_number + "-2",
            "sn": phone_number + "-2",
            "AstApplication": "Voicemail",
            "AstContext": "internal",
            "AstPriority": "2",
            "AstExtension": phone_number,
            "objectClass": ("top", "inetOrgPerson", "AsteriskExtension")
        }

        modlist = ldap.modlist.addModlist(entry)
        self.l.add_s(dn, modlist)

    def del_group_extension(self, phone_number):
        """
        Del an asterisk extension entry and
        a asterisk voicemail entry in directory

        @param phone_number: user phone number
        @type phone_number: str
        """

        dn = "cn=" + phone_number + "-1, " + self.extensionsDN
        self.delRecursiveEntry(dn)

        dn = "cn=" + phone_number + "-2, " + self.extensionsDN
        self.delRecursiveEntry(dn)


class TrunkSIPConfig():

    def get_sip_config(self):
        """
        Return the content of the file that contain trunk SIP configuration
        """
        with open(TRUNK_SIP_CONFIG_FILE, 'a+') as f:
            sip_config = f.readlines()

        return ''.join(sip_config)

    def update_sip_config(self, sip_config):
        """
        Update the trunk SIP configuration file

        @param sip_config: trunk sip configuration
        @type sup_config: string
        """
        with open(SIP_CONFIG_FILE, 'a+') as f:
            lines = f.readlines()
            last_line = lines[-1].decode('utf8')

            if not last_line == u'#include {}'.format(TRUNK_SIP_CONFIG_FILE):
                f.write(u'#include {}'.format(TRUNK_SIP_CONFIG_FILE).encode('utf8'))

        with open(TRUNK_SIP_CONFIG_FILE, 'w') as f:
            for line in sip_config.split('\n'):
                f.write(u'{}\n'.format(line).encode('utf8'))

        self._chown_config_file(TRUNK_SIP_CONFIG_FILE)

    def get_dialplan_config(self):
        """
        Return the content of the file that contain the dialplan configuration
        """
        with open(DIALPLAN_CONFIG_FILE, 'a+') as f:
            dialplan_config = f.readlines()

        return ''.join(dialplan_config)

    def update_dialplan_config(self, dialplan_config):
        """
        Update the dialplan configuration file

        @param dialplan_config: dialplan configuration
        @type dialplan_config: string
        """
        with open(EXTENSIONS_CONFIG_FILE, 'a+') as f:
            lines = f.readlines()
            last_line = lines[-1].decode('utf8')

            if not last_line == u'#include {}'.format(DIALPLAN_CONFIG_FILE):
                f.write(u'#include {}'.format(DIALPLAN_CONFIG_FILE).encode('utf8'))

        with open(DIALPLAN_CONFIG_FILE, 'w') as f:
            for line in dialplan_config.split('\n'):
                f.write(u'{}\n'.format(line).encode('utf8'))

        self._chown_config_file(DIALPLAN_CONFIG_FILE)

    def _chown_config_file(self, file):
        """
        Change the owner/group of the file

        @param file: configuration file to be chowned
        @type file: string
        """
        user_id = pwd.getpwnam('asterisk').pw_uid
        group_id = grp.getgrnam('asterisk').gr_gid

        os.chown(file, user_id, group_id)
